// Copyright (c) 2017-2024, Lawrence Livermore National Security, LLC and
// other Axom Project Developers. See the top-level LICENSE file for details.
//
// SPDX-License-Identifier: (BSD-3-Clause)

#ifndef AXOM_ANNOTATION_MACROS_HPP_
#define AXOM_ANNOTATION_MACROS_HPP_

#include "axom/config.hpp"
#include "axom/core/utilities/Annotations.hpp"

/*!
 * \def AXOM_ANNOTATE_BEGIN( name )
 * 
 * \brief This macro is used to mark the beginning of an annotated region
 * \note \a AXOM_ANNOTATE_BEGIN must be closed with a matching \a AXOM_ANNOTATE_END
 */
#define AXOM_ANNOTATE_BEGIN(name) axom::utilities::annotations::begin(name)

/*!
 * \def AXOM_ANNOTATE_END( name )
 * 
 * \brief This macro is used to mark the end of an annotated region
 * \note \a AXOM_ANNOTATE_END must close a matching \a AXOM_ANNOTATE_BEGIN
 */
#define AXOM_ANNOTATE_END(name) axom::utilities::annotations::end(name)

/*!
 * \def AXOM_ANNOTATE_SCOPE( name )
 * \brief This macro is used to annotate a region within an enclosing scope
 * and is automatically closed when its enclosing scope ends
 * 
 * \warning The \a AXOM_ANNOTATE_SCOPE can only be called once within a given scope
 * 
 * \note \a AXOM_ANNOTATE_SCOPE only annotates the portion of the scope following this macro. 
 * It adds a variable whose constructor begins an annotation region and whose destructor, 
 * which is called at the end of the scope, ends the annotation region.
 */
#ifdef AXOM_USE_CALIPER
  #define AXOM_ANNOTATE_SCOPE(name)                        \
    cali::Annotation::Guard cali_autogenerated_guard_name( \
      cali::Annotation("region").begin(std::string(name).c_str()))
#else
  #define AXOM_ANNOTATE_SCOPE(name) \
    do                              \
    {                               \
    } while(false)
#endif

/*!
 * \def AXOM_ANNOTATE_METADATA( name, value, category )
 * \brief This macro is used to add metadata to the run and is added to caliper output
 * in modes that generate an output file.
 * \see axom::utilities::annotations::declare_metadata
 */
#define AXOM_ANNOTATE_METADATA(name, value, category) \
  axom::utilities::annotations::declare_metadata(name, value, category)

#endif  // AXOM_ANNOTATION_MACROS_HPP_
