# Copyright (c) 2017-2024, Lawrence Livermore National Security, LLC and
# other Axom Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: (BSD-3-Clause)
#------------------------------------------------------------------------------

cmake_minimum_required(VERSION 3.8 FATAL_ERROR)


####### Expanded from @PACKAGE_INIT@ by configure_package_config_file() #######
####### Any changes to this file will be overwritten by the next CMake run ####
####### The input file was axom-config.cmake.in                            ########

get_filename_component(PACKAGE_PREFIX_DIR "${CMAKE_CURRENT_LIST_DIR}/../" ABSOLUTE)

macro(set_and_check _var _file)
  set(${_var} "${_file}")
  if(NOT EXISTS "${_file}")
    message(FATAL_ERROR "File or directory ${_file} referenced by variable ${_var} does not exist !")
  endif()
endmacro()

macro(check_required_components _NAME)
  foreach(comp ${${_NAME}_FIND_COMPONENTS})
    if(NOT ${_NAME}_${comp}_FOUND)
      if(${_NAME}_FIND_REQUIRED_${comp})
        set(${_NAME}_FOUND FALSE)
      endif()
    endif()
  endforeach()
endmacro()

####################################################################################

if(NOT AXOM_FOUND)

  #----------------------------------------------------------------------------
  # Set version and paths
  #----------------------------------------------------------------------------
  
  set(AXOM_VERSION          "v0.10.1")
  set(AXOM_VERSION_MAJOR    "0")
  set(AXOM_VERSION_MINOR    "10")
  set(AXOM_VERSION_PATCH    "1")
  
  set(AXOM_GIT_SHA          "")

  set(AXOM_INSTALL_PREFIX   "/usr/local")
  set(AXOM_INCLUDE_DIRS     "${AXOM_INSTALL_PREFIX}/include")
  
  #----------------------------------------------------------------------------
  # Set user configuration options and features
  #----------------------------------------------------------------------------

  # Language features
  set(AXOM_ENABLE_FORTRAN     "OFF")
  set(AXOM_USE_CUDA           "")
  set(AXOM_USE_HIP            "")
  set(AXOM_USE_MPI            "TRUE")
  set(AXOM_USE_OPENMP         "")

  # Axom components
  set(AXOM_COMPONENTS_ENABLED core;lumberjack;slic;slam;primal;sidre;mint;spin;inlet;klee;quest;multimat)

  set(AXOM_ENABLE_INLET       "ON")
  set(AXOM_ENABLE_KLEE        "ON")
  set(AXOM_ENABLE_LUMBERJACK  "ON")
  set(AXOM_ENABLE_MINT        "ON")
  set(AXOM_ENABLE_PRIMAL      "ON")
  set(AXOM_ENABLE_QUEST       "ON")
  set(AXOM_ENABLE_SIDRE       "ON")
  set(AXOM_ENABLE_SLAM        "ON")
  set(AXOM_ENABLE_SLIC        "ON")
  set(AXOM_ENABLE_SPIN        "ON")

  # Axom built-in TPLs
  set(AXOM_USE_CLI11          "TRUE")
  set(AXOM_USE_FMT            "TRUE")
  set(AXOM_USE_SOL            "TRUE")
  set(AXOM_USE_SPARSEHASH     "TRUE")

  # Axom external TPLs
  set(AXOM_USE_ADIAK          "")
  set(AXOM_USE_C2C            "")
  set(AXOM_USE_CALIPER        "")
  set(AXOM_USE_CAMP           "")
  set(AXOM_USE_CONDUIT        "TRUE")
  set(AXOM_USE_HDF5           "TRUE")
  set(AXOM_USE_LUA            "TRUE")
  set(AXOM_USE_MFEM           "TRUE")
  set(AXOM_USE_RAJA           "")
  set(AXOM_USE_SCR            "")
  set(AXOM_USE_UMPIRE         "")

  # Configration for Axom compiler defines
  set(AXOM_DEBUG_DEFINE         "DEFAULT")
  set(AXOM_DEBUG_DEFINE_STRING  "")

  #----------------------------------------------------------------------------
  # Bring in required dependencies for this axom configuration
  #----------------------------------------------------------------------------
  include(CMakeFindDependencyMacro)

  # Enable various find commands to look in non-default paths
  set_property(GLOBAL PROPERTY FIND_LIBRARY_USE_LIB32_PATHS TRUE)
  set_property(GLOBAL PROPERTY FIND_LIBRARY_USE_LIB64_PATHS TRUE)
  set_property(GLOBAL PROPERTY FIND_LIBRARY_USE_LIBX32_PATHS TRUE)

  # adiak
  if(AXOM_USE_ADIAK)
    set(AXOM_ADIAK_DIR     "")
    if(NOT ADIAK_DIR) 
      set(ADIAK_DIR ${AXOM_ADIAK_DIR})
    endif()
    find_dependency(adiak REQUIRED 
                    PATHS "${ADIAK_DIR}"
                          "${ADIAK_DIR}/lib/cmake/adiak"
                    NO_SYSTEM_ENVIRONMENT_PATH)
  endif()

  # c2c
  if(AXOM_USE_C2C)
    set(AXOM_C2C_DIR     "")
    # Note: Targets not currently imported
  endif()

  # caliper
  if(AXOM_USE_CALIPER)
    set(AXOM_CALIPER_DIR   "")
    if(NOT CALIPER_DIR) 
      set(CALIPER_DIR ${AXOM_CALIPER_DIR})
    endif()
    find_dependency(caliper REQUIRED 
                    PATHS "${CALIPER_DIR}"
                          "${CALIPER_DIR}/share/cmake/caliper"
                    NO_SYSTEM_ENVIRONMENT_PATH)
  endif()

  # camp
  if(AXOM_USE_CAMP)
    set(AXOM_CAMP_DIR      "")
    if(NOT CAMP_DIR)
      set(CAMP_DIR ${AXOM_CAMP_DIR})
    endif()
    find_dependency(camp REQUIRED PATHS "${CAMP_DIR}" NO_SYSTEM_ENVIRONMENT_PATH)
  endif()

  # umpire
  if(AXOM_USE_UMPIRE)
    set(AXOM_UMPIRE_DIR   "")
    if(NOT UMPIRE_DIR)
      set(UMPIRE_DIR ${AXOM_UMPIRE_DIR})
    endif()
    find_dependency(umpire REQUIRED PATHS "${UMPIRE_DIR}" NO_SYSTEM_ENVIRONMENT_PATH)
  endif()

  # raja
  if(AXOM_USE_RAJA)
    set(AXOM_RAJA_DIR     "")
    if(NOT RAJA_DIR)
      set(RAJA_DIR ${AXOM_RAJA_DIR})
    endif()
    find_dependency(RAJA REQUIRED PATHS "${RAJA_DIR}" NO_SYSTEM_ENVIRONMENT_PATH)
  endif()

  # conduit
  if(AXOM_USE_CONDUIT)
    set(AXOM_CONDUIT_DIR  "/usr/local")
    if(NOT CONDUIT_DIR) 
      set(CONDUIT_DIR ${AXOM_CONDUIT_DIR}) 
    endif()

    # Load mpi targets because we require the optional Conduit mpi targets
    if(AXOM_USE_MPI)
      find_package(MPI REQUIRED)
    endif()

    find_dependency(Conduit REQUIRED
                    PATHS "${CONDUIT_DIR}"
                          "${CONDUIT_DIR}/lib/cmake/conduit"
                    NO_SYSTEM_ENVIRONMENT_PATH)
  endif()

  # hdf5
  if(AXOM_USE_HDF5)
    set(AXOM_HDF5_DIR     "/usr/local")
    # Note: 'hdf5" target is exported as 'axom::hdf5'
  endif()

  # lua
  if(AXOM_USE_LUA)
    set(AXOM_LUA_DIR     "/usr/local")
    # Note: 'lua" target is exported as 'axom::lua'
  endif()

  # mfem
  if(AXOM_USE_MFEM)
    set(AXOM_MFEM_DIR     "/usr/local")
    # Note: 'mfem" target is exported as 'axom::mfem'
  endif()

  # scr
  if(AXOM_USE_SCR)
    set(AXOM_SCR_DIR      "")
    # Note: 'scr" target is exported as 'axom::scr'
  endif()

  #----------------------------------------------------------------------------
  # Include targets exported by cmake
  #----------------------------------------------------------------------------
  get_filename_component(AXOM_CMAKE_CONFIG_DIR "${CMAKE_CURRENT_LIST_FILE}" PATH)
  get_filename_component(_IMPORT_PREFIX "${AXOM_CMAKE_CONFIG_DIR}" PATH)

  if(_IMPORT_PREFIX STREQUAL "/")
    set(_IMPORT_PREFIX "")
  endif()

  # we want the import root, which is right above the "lib" prefix
  get_filename_component(_IMPORT_ROOT "${_IMPORT_PREFIX}" PATH)

  include(${AXOM_CMAKE_CONFIG_DIR}/BLTSetupTargets.cmake)
  include(${AXOM_CMAKE_CONFIG_DIR}/axom-targets.cmake)

  # Create convenience target that bundles all Axom targets (axom)
  add_library(axom INTERFACE IMPORTED)

  set_property(TARGET axom
               APPEND PROPERTY
               INTERFACE_INCLUDE_DIRECTORIES "${_IMPORT_ROOT}/include/")

  foreach(_component ${AXOM_COMPONENTS_ENABLED})
    set_property(TARGET axom
                 PROPERTY INTERFACE_LINK_LIBRARIES
                 axom::${_component}
                 APPEND)
  endforeach()

  #---------------------------------------------------------------------------
  # Remove non-existant INTERFACE_INCLUDE_DIRECTORIES from imported targets
  # to work around CMake error
  #---------------------------------------------------------------------------

  # Support IN_LIST operator for if()
  # Policy added in 3.3+
  if(POLICY CMP0057)
    cmake_policy(SET CMP0057 NEW)
  endif()

  ## axom_find_target_dependencies(TARGET <target> TLIST <list name>)
  ##
  ## Store all targets' dependencies (link libraries and interface link libraries)
  ## recursively in the variable name TLIST holds.
  macro(axom_find_target_dependencies)

      set(options)
      set(singleValuedArgs TARGET TLIST)
      set(multiValuedArgs)

      # parse the arguments to the macro
      cmake_parse_arguments(arg
           "${options}" "${singleValuedArgs}" "${multiValuedArgs}" ${ARGN})

      # check for required arguments
      if(NOT DEFINED arg_TARGET)
          message(FATAL_ERROR "TARGET is a required parameter for the axom_find_target_dependencies macro")
      endif()

      if(NOT DEFINED arg_TLIST OR NOT DEFINED ${arg_TLIST})
          message(FATAL_ERROR "TLIST is a required parameter for the axom_find_target_dependencies macro")
      endif()

      set(_depends_on "")

      # Get dependencies if BLT registered library
      string(TOUPPER ${arg_TARGET} _target_upper)
      if(_BLT_${_target_upper}_IS_REGISTERED_LIBRARY)
          list(APPEND _depends_on "${_BLT_${_target_upper}_DEPENDS_ON}")
      endif()

      # Get dependencies if CMake target
      if(TARGET ${arg_TARGET})
          get_property(_target_type TARGET ${arg_TARGET} PROPERTY TYPE)
          if(NOT "${_target_type}" STREQUAL "INTERFACE_LIBRARY")
              get_target_property(_propval ${arg_TARGET} LINK_LIBRARIES)
              if(_propval)
                  list(APPEND _depends_on ${_propval})
              endif()
          endif()

          # get interface link libraries
          get_target_property(_propval ${arg_TARGET} INTERFACE_LINK_LIBRARIES)
          if (_propval)
              list(APPEND _depends_on ${_propval})
          endif()
      endif()

      blt_list_remove_duplicates(TO _depends_on)
      foreach(t ${_depends_on})
          if (NOT "${t}" IN_LIST ${arg_TLIST})
              list(APPEND ${arg_TLIST} ${t})
              axom_find_target_dependencies(TARGET ${t} TLIST ${arg_TLIST})
          endif()
      endforeach()

      unset(_depends_on)
  endmacro(axom_find_target_dependencies)

  set(_deps "")
  axom_find_target_dependencies(TARGET axom TLIST _deps)
  list(REMOVE_DUPLICATES _deps)
  message(STATUS "Removing non-existant include directories from Axom's dependencies...")

  foreach(_target ${_deps})
    if(TARGET ${_target})
      get_target_property(_dirs ${_target} INTERFACE_INCLUDE_DIRECTORIES)
      set(_existing_dirs)
      foreach(_dir ${_dirs})
        if (EXISTS "${_dir}")
          list(APPEND _existing_dirs "${_dir}")
        endif()
      endforeach()
      if (_existing_dirs)
        set_target_properties(${_target} PROPERTIES
                              INTERFACE_INCLUDE_DIRECTORIES "${_existing_dirs}" )
      endif()
    endif()
  endforeach()

  #----------------------------------------------------------------------------
  # Indicate that Axom is correctly set up
  #----------------------------------------------------------------------------
  set(AXOM_FOUND TRUE)

endif()
